// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2018 Verisilicon Inc.
 *
 * This is misc ip management driver for Linux.
 * There are eight types of small ip, we call them misc ip.they are: L2CACH_VCD,
 * L2CACH_VCE, DEC400_VCD, DEC400_VCE, TCACHE, DTRC, DEC, L2CACHE_DTRC_TCACH
 * and PLL_CONTROLLER. This file provide register operation and initialization,
 * like read/write a register or pull/push a batch of registers, because the
 * codec IPs(vc8000d, vc8000e, bigsea) have the similar interfaces, they use
 * unified register operation APIs.
 */

#include <linux/pci.h>
#include <linux/pagemap.h>
#include <linux/delay.h>
#include <linux/timer.h>

#include "common.h"
#include "dec_ip.h"
#include "transcoder.h"
#include "regs.h"

#define HOST_DEC_IOSIZE 0x200000

struct ip_info {
	u32 base_off;
	void __iomem *hwregs;
	u32 iosize;
	u32 *shadow;
};

struct dec_ip {
	struct ip_info host_dec;
	struct sn_tranx_t *tdev;
};

u32 host_dec_info[2] = { HOST_DEC_OFF, HOST_DEC_IOSIZE };

static void check_core_status(struct sn_tranx_t *tdev, int core_id,
			      char *fun_name, int reg_id)
{
}

/* write one register. */
static int write_one_reg(struct core_desc core, struct sn_tranx_t *tdev)
{
	int ret, id;
	u32 val;
	struct dec_ip *tdec = tdev->modules[SN_MODULE_HOST_DEC];

	if (tdev->hw_err_flag)
		return tdev->hw_err_flag;

	check_core_status(tdev, core.id, "write_one_reg", core.reg_id);
	id = core.reg_id;
	ret = copy_from_user(&val, core.regs, 4);
	if (ret)
		sn_pri(tdev, SN_ERR, "host_dec_ip: %s failed.\n", __func__);
	else {
		writel(val, tdec->host_dec.hwregs + id * 4);
	}

	return ret ? -EFAULT : 0;
}

/* read one register. */
static int read_one_reg(struct core_desc core, struct sn_tranx_t *tdev)
{
	int ret, id;
	u32 val;
	struct dec_ip *tdec = tdev->modules[SN_MODULE_HOST_DEC];

	if (tdev->hw_err_flag)
		return tdev->hw_err_flag;

	check_core_status(tdev, core.id, "read_one_reg", core.reg_id);
	id = core.reg_id;
	val = readl(tdec->host_dec.hwregs + id * 4);

	/* put registers to user space */
	ret = copy_to_user(core.regs, &val, 4);
	if (ret)
		sn_pri(tdev, SN_ERR, "host_dec_ip: %s failed.\n", __func__);

	return ret ? -EFAULT : 0;
}

/*
 * get value from userspace, then batch write registers.
 * start id is core->reg_id, size is core->size.
 */
static int regs_batch_write(struct core_desc core, struct sn_tranx_t *tdev)
{
	int ret, i, regs_cnt;
	struct dec_ip *tdec = tdev->modules[SN_MODULE_HOST_DEC];
	struct reg_desc *regs_info = (struct reg_desc *)tdec->host_dec.shadow;

	if (tdev->hw_err_flag)
		return tdev->hw_err_flag;

	check_core_status(tdev, core.id, "regs_batch_write", core.reg_id);
	regs_cnt = core.size / sizeof(struct reg_desc);
	ret = copy_from_user(regs_info, core.regs, core.size);
	if (ret)
		sn_pri(tdev, SN_ERR, "host_dec_ip: %s failed.\n", __func__);
	else {
		for (i = 0; i < regs_cnt; i++) {
			writel(regs_info[i].val,
			       tdec->host_dec.hwregs + regs_info[i].id * 4);
		}
	}

	return ret ? -EFAULT : 0;
}

/*
 * batch read registers, then copy them to userspace.
 * start id is core->reg_id, size is core->size.
 */
static int regs_batch_read(struct core_desc core, struct sn_tranx_t *tdev)
{
	int ret, i, regs_cnt;
	struct dec_ip *tdec = tdev->modules[SN_MODULE_HOST_DEC];
	struct reg_desc *regs_info = (struct reg_desc *)tdec->host_dec.shadow;

	if (tdev->hw_err_flag)
		return tdev->hw_err_flag;

	check_core_status(tdev, core.id, "regs_batch_read", core.reg_id);
	ret = copy_from_user(regs_info, core.regs, core.size);
	if (ret) {
		sn_pri(tdev, SN_ERR, "host_dec_ip: %s failed.\n", __func__);
		return -EFAULT;
	}
	regs_cnt = core.size / sizeof(struct reg_desc);
	for (i = 0; i < regs_cnt; i++) {
		regs_info[i].val =
			readl(tdec->host_dec.hwregs + regs_info[i].id * 4);
	}
	ret = copy_to_user(core.regs, regs_info, core.size);
	if (ret)
		sn_pri(tdev, SN_ERR, "host_dec_ip: %s failed.\n", __func__);

	return ret ? -EFAULT : 0;
}

int host_dec_init(struct sn_tranx_t *tdev)
{
	struct dec_ip *tdec;
	void *tmp_mem;

	tdec = kzalloc(sizeof(struct dec_ip), GFP_KERNEL);
	if (!tdec) {
		sn_pri(tdev, SN_ERR,
		       "host_dec_ip: %s kmalloc bigseacodec failed\n",
		       __func__);
		goto out;
	}
	tdev->modules[SN_MODULE_HOST_DEC] = tdec;
	tdec->tdev = tdev;

	tmp_mem = vzalloc(HOST_DEC_IOSIZE);
	if (!tmp_mem) {
		sn_pri(tdev, SN_ERR, "host_dec_ip: malloc regs mem failed.\n");
		goto out_free_dev;
	}

	tdec->host_dec.base_off = host_dec_info[0];
	tdec->host_dec.hwregs = tdev->bar2_virt + host_dec_info[0];
	tdec->host_dec.iosize = host_dec_info[1];
	tdec->host_dec.shadow = tmp_mem;

	sn_pri(tdev, SN_INF, "host_dec_ip: module initialize done.\n");
	return 0;

out_free_dev:
	kfree(tdec);
out:
	sn_pri(tdec->tdev, SN_ERR, "host_dec_ip: module initialize failed.\n");
	return -EFAULT;
}

int host_dec_release(struct sn_tranx_t *tdev)
{
	struct dec_ip *tdec = tdev->modules[SN_MODULE_HOST_DEC];

	kfree(tdec);
	sn_pri(tdev, SN_DBG, "host_dec_ip: remove module done.\n");
	return 0;
}

long host_dec_ioctl(struct file *filp, unsigned int cmd, unsigned long arg,
		    struct sn_tranx_t *tdev)
{
	int ret = 0;
	struct ip_desc ipdesc;
	void __user *argp = (void __user *)arg;

	if (copy_from_user(&ipdesc, argp, sizeof(struct ip_desc))) {
		sn_pri(tdev, SN_ERR, "host_dec_ip: %s copy from user failed\n",
		       __func__);
		return -EFAULT;
	}

	switch (cmd) {
	case SN_TRANX_RD_REG:
		ret = read_one_reg(ipdesc.core, tdev);
		break;
	case SN_TRANX_WR_REG:
		ret = write_one_reg(ipdesc.core, tdev);
		break;
	case SN_TRANX_PULL_REGS:
		ret = regs_batch_read(ipdesc.core, tdev);
		break;
	case SN_TRANX_PUSH_REGS:
		ret = regs_batch_write(ipdesc.core, tdev);
		break;
	default:
		sn_pri(tdev, SN_ERR, "host_dec_ip: %s, cmd:0x%x error.\n",
		       __func__, cmd);
		ret = -EINVAL;
	}

	return ret;
}
